#!/bin/bash

i=$1
shift

if [ -z "$i" ]; then
    echo "Usage:"
    echo "      $0 <interface>"
    exit 1
fi

WINDRIVER=0
if (grep -qiE "Wind River" /etc/issue /etc/*release* 2>/dev/null); then
    WINDRIVER=1
fi

OPENIBD_CONFIG=${OPENIBD_CONFIG:-"/etc/infiniband/openib.conf"}
CONFIG=$OPENIBD_CONFIG
export LANG=en_US.UTF-8

if [ ! -f $CONFIG ]; then
    echo No InfiniBand configuration found
    exit 0
fi

. $CONFIG
IPOIB_MTU=${IPOIB_MTU:-65520}

# set configuration files
conf_files=/etc/network/interfaces
if (grep -w source /etc/network/interfaces 2>/dev/null | grep -qvE "^\s*#" 2>/dev/null); then
    # get absolute file paths
    for line in $(grep -w source /etc/network/interfaces 2>/dev/null | grep -vE "^\s*#" 2>/dev/null);
    do
        ff=$(echo "$line" | awk '{print $NF}')

        # check if it's absolute path
        if [ -f "$ff" ]; then
            conf_files="$conf_files $ff"
            continue
        fi

        # check if it's relative path
        if [ -z "$(ls $ff 2>/dev/null)" ]; then
            ff="/etc/network/$ff"
        fi

        # support wildcards
        for file in $(ls -1 $ff 2>/dev/null)
        do
            if [ -f "$file" ]; then
                conf_files="$conf_files $file"
            fi
        done
    done
fi

log_msg()
{
    logger -t 'mlnx_interface_mgr' -i "$@"
}

set_ipoib_cm()
{
    local i=$1
    shift
    local mtu=$1
    shift
    local is_up=""

    if [ ! -e /sys/class/net/${i}/mode ]; then
        log_msg "Failed to configure IPoIB connected mode for ${i}"
        return 1
    fi

    mtu=${mtu:-$IPOIB_MTU}

    #check what was the previous state of the interface
    is_up=`/sbin/ip link show $i | grep -w UP`

    /sbin/ip link set ${i} down
    if [ $? -ne 0 ]; then
        log_msg "set_ipoib_cm: Failed to bring down ${i} in order to change connection mode"
        return 1
    fi

    if [ -w /sys/class/net/${i}/mode ]; then
        echo connected > /sys/class/net/${i}/mode
        if [ $? -eq 0 ]; then
            log_msg "set_ipoib_cm: ${i} connection mode set to connected"
        else
            log_msg "set_ipoib_cm: Failed to change connection mode for ${i} to connected"
            RC=1
        fi
    else
        log_msg "set_ipoib_cm: cannot write to /sys/class/net/${i}/mode"
        RC=1
    fi
    /sbin/ip link set ${i} mtu ${mtu}
    if [ $? -ne 0 ]; then
        log_msg "set_ipoib_cm: Failed to set mtu for ${i}"
        RC=1
    fi

    #if the intf was up returns it to
    if [ -n "$is_up" ]; then
        /sbin/ip link set ${i} up
        if [ $? -ne 0 ]; then
            log_msg "set_ipoib_cm: Failed to bring up ${i} after setting connection mode to connected"
            RC=1
        fi
    fi

    return $RC
}

set_RPS_cpu()
{
    local i=$1
    shift

    # Silently ignore Pkeys
    if [ ! -e /sys/class/net/${i}/device ]; then
        return 0
    fi

    if [ ! -e /sys/class/net/${i}/queues/rx-0/rps_cpus ]; then
        log_msg "set_RPS_cpu: Failed to configure RPS cpu for ${i}"
        return 1
    fi

    cat /sys/class/net/${i}/device/local_cpus >  /sys/class/net/${i}/queues/rx-0/rps_cpus
    if [ $? -eq 0 ]; then
        log_msg "set_RPS_cpu: Configured RPS cpu for ${i}"
    else
        log_msg "set_RPS_cpu: Failed to configure RPS cpu for ${i}"
        return 1
    fi

    return 0
}

bring_up()
{
    local i=$1
    shift
    local RC=0

    if ! (grep -wh ${i} $conf_files 2>/dev/null | grep -qvE "^\s*#" 2>/dev/null); then
        log_msg "No configuration found for ${i}"
        return 4
    fi

    # get current interface status
    local is_up=""
    is_up=`/sbin/ip link show $i | grep -w UP`
    if [ -z "$is_up" ]; then
        if [ $WINDRIVER -eq 0 ]; then
            /sbin/ifup --force ${i} > /dev/null 2>&1
        else
            env PATH=$PATH:/sbin /sbin/ifup -f ${i} 2>&1
        fi
        if [ $? -eq 0 ]; then
            log_msg "Bringing up interface $i: PASSED"
        else
            log_msg "Bringing up interface $i: FAILED"
            return 1
        fi
    fi

    MTU=`/usr/sbin/net-interfaces get-mtu ${i}`

    # relevant for IPoIB interfaces only
    if (/sbin/ethtool -i ${i} 2>/dev/null | grep -q ib_ipoib); then
        if [ "X${SET_IPOIB_CM}" == "Xyes" ]; then
            set_ipoib_cm ${i} ${MTU}
            if [ $? -ne 0 ]; then
                RC=1
            fi
        elif [ "X${SET_IPOIB_CM}" == "Xauto" ]; then
            # handle mlx5 interfaces, assumption: mlx5 interface will be with CM mode.
            if [ "X$(basename `readlink -f /sys/class/net/${i}/device/driver/module 2>/dev/null` 2>/dev/null)" == "Xmlx5_core" ]; then
                set_ipoib_cm ${i} ${MTU}
                if [ $? -ne 0 ]; then
                    RC=1
                fi
            fi
        fi
        # Spread the one and only RX queue to more CPUs using RPS.
        local num_rx_queue=$(ls -l /sys/class/net/${i}/queues/ 2>/dev/null | grep rx-  | wc -l | awk '{print $1}')
        if [ $num_rx_queue -eq 1 ]; then
            set_RPS_cpu ${i}
            if [ $? -ne 0 ]; then
                RC=1
            fi
        fi
    fi

    bond=`/usr/sbin/net-interfaces get-bond-master ${i}`
    if [ ! -z "$bond" ]; then
        /sbin/ifenslave -f $bond ${i} > /dev/null 2>&1
        if [ $? -ne 0 ]; then
            log_msg "$i - briging up bond master $MASTER: PASSED"
        else
            log_msg "$i - briging up bond master $MASTER: FAILED"
            RC=1
        fi
    fi

    return $RC
}

# main
log_msg "Setting up Mellanox network interface: $i"

# bring up the interface
bring_up $i
if [ $? -eq 1 ]; then
    log_msg "Couldn't fully configure ${i}, review system logs and restart network service after fixing the issues."
fi

# Bring up child interfaces if configured.
for file in $conf_files
do
    while read _line
    do
        if [[ ! "$_line" =~ ^# && "$_line" =~ $i\.[0-9]* && "$_line" =~ "iface" ]]
        then
            ifname=$(echo $_line | cut -f2 -d" ")

            if [ ! -f /sys/class/net/$i/create_child ]; then
                continue
            fi

            suffix=$(echo ${ifname##*.} | tr '[:upper:]' '[:lower:]')
            if [[ ${suffix} =~ ^[0-9a-f]{1,4}$ ]]; then
                hexa=$(printf "%x" $(( 0x${suffix} | 0x8000 )))
                if [[ ${hexa} != ${suffix} ]]; then
                    log_msg "Error: MSB is NOT set for pkey ${suffix} (should be ${hexa}); skipping interface ${ifname}."
                    continue
                fi
            else
                log_msg "Error: pkey ${suffix} is not hexadecimal (maximum 4 digits); skipping."
                continue
            fi
            pkey=0x${hexa}

            if [ ! -e /sys/class/net/$ifname ] ; then
                {
                local retry_cnt=0
                echo $pkey > /sys/class/net/$i/create_child
                while [[ $? -ne 0 && $retry_cnt -lt 10 ]]; do
                    sleep 1
                    let retry_cnt++
                    echo $pkey > /sys/class/net/$i/create_child
                done
                } > /dev/null 2>&1
            fi

            bring_up $ifname
            if [ $? -eq 1 ]; then
                log_msg "Couldn't fully configure ${ifname}, review system logs and restart network service after fixing the issues."
            fi
        fi
    done < "$file"
done
